@@ -27,6 +27,7 @@
 #include "userlist-gui.h"
 #include "../common/xchatc.h"
 #include "../common/outbound.h"
+#include "../common/fe.h"
 #include "gui.h"
 #include "channel-list.h"
 #include "preferences.h"
@@ -36,8 +37,9 @@
 #include "util.h"
 #include "preferences-dialog.h"
 
-#ifdef HAVE_GTKSPELL
-#include <gtkspell/gtkspell.h>
+#ifdef HAVE_LIBSEXY
+#include <libsexy/sexy-url-label.h>
+#include <libsexy/sexy-spell-entry.h>
 #endif
 
 #include <sys/types.h>
@@ -54,8 +56,6 @@ static void *last_search_position = NULL;
 
 static void on_main_window_close (GtkWidget *widget, GdkEvent *event, gpointer data);
 static void on_discussion_jump_activate (GtkAccelGroup *accelgroup, GObject *arg1, guint arg2, GdkModifierType arg3, gpointer data);
-static void on_discussion_plus_activate (GtkAccelGroup *accelgroup, GObject *arg1, guint arg2, GdkModifierType arg3, gpointer data);
-static void on_discussion_minus_activate (GtkAccelGroup *accelgroup, GObject *arg1, guint arg2, GdkModifierType arg3, gpointer data);
 static void on_pgup (GtkAccelGroup *accelgroup, GObject *arg1, guint arg2, GdkModifierType arg3, gpointer data);
 static void on_pgdn (GtkAccelGroup *accelgroup, GObject *arg1, guint arg2, GdkModifierType arg3, gpointer data);
 
@@ -68,7 +68,6 @@ static void on_edit_copy_activate (GtkAction *action, gpointer data);
 static void on_edit_paste_activate (GtkAction *action, gpointer data);
 static void on_edit_clear_activate (GtkAction *action, gpointer data);
 static void on_edit_preferences_activate (GtkAction *action, gpointer data);
-static void on_insert_color_code_activate (GtkAction *action, gpointer data);
 static void on_network_information_activate (GtkAction *action, gpointer data);
 static void on_network_reconnect_activate (GtkAction *action, gpointer data);
 static void on_network_disconnect_activate (GtkAction *action, gpointer data);
@@ -88,12 +87,11 @@ static void on_go_next_network_activate (GtkAction *action, gpointer data);
 static void on_go_previous_discussion_activate (GtkAction *action, gpointer data);
 static void on_go_next_discussion_activate (GtkAction *action, gpointer data);
 static void on_help_about_activate (GtkAction *action, gpointer data);
+static void on_nickname_clicked (GtkButton *widget, gpointer user_data);
 
 static void on_add_widget (GtkUIManager *manager, GtkWidget *menu, GtkWidget *menu_vbox);
 
-#if (GTK_CHECK_VERSION(2,5,0))
 static void on_expand_topic (GtkExpander *expander, gpointer data);
-#endif
 
 static void on_text_entry_activate (GtkWidget *widget, gpointer data);
 static gboolean on_text_entry_key (GtkWidget *widget, GdkEventKey *key, gpointer data);
@@ -117,11 +115,7 @@ static GtkActionEntry action_entries [] = {
 	{ "PopupAction", NULL, _("") },
 
 	/* IRC menu */
-#if (GTK_CHECK_VERSION(2,5,0))
 	{ "IRCConnect", GTK_STOCK_CONNECT, _("_Connect"), "<control>N", NULL, G_CALLBACK (on_irc_connect_activate) },
-#else
-	{ "IRCConnect", NULL, _("_Connect"), "<control>N", NULL, G_CALLBACK (on_irc_connect_activate) },
-#endif
 	{ "IRCDownloads", NULL, _("_File Transfers"), "<alt>F", NULL, G_CALLBACK (on_irc_downloads_activate) },
 	{ "IRCQuit", GTK_STOCK_QUIT, _("_Quit"), "<control>Q", NULL, G_CALLBACK (on_irc_quit_activate) },
 
@@ -132,21 +126,18 @@ static GtkActionEntry action_entries [] = {
 	{ "EditClear", GTK_STOCK_CLEAR, _("C_lear"), "", NULL, G_CALLBACK (on_edit_clear_activate) },
 	{ "EditPreferences", GTK_STOCK_PREFERENCES, _("Prefere_nces"), "", NULL, G_CALLBACK (on_edit_preferences_activate) },
 
-	/* Insert menu */
-	{ "InsertColorCode", GTK_STOCK_SELECT_COLOR, _("Color Code..."), "", NULL, G_CALLBACK (on_insert_color_code_activate) },
-
 	/* Network menu */
 	{ "NetworkInformation", GTK_STOCK_DIALOG_INFO, _("_Information"), "", NULL, G_CALLBACK (on_network_information_activate) },
 	{ "NetworkReconnect", GTK_STOCK_REFRESH, _("_Reconnect"), "<control>R", NULL, G_CALLBACK (on_network_reconnect_activate) },
 	{ "NetworkDisconnect", GTK_STOCK_STOP, _("_Disconnect"), "", NULL, G_CALLBACK (on_network_disconnect_activate) },
-	{ "NetworkClose", GTK_STOCK_CLOSE, _("_Close"), "", NULL, G_CALLBACK (on_network_close_activate) },
+	{ "NetworkClose", GTK_STOCK_CLOSE, _("_Close"), "<shift><control>W", NULL, G_CALLBACK (on_network_close_activate) },
 	{ "NetworkChannels", GTK_STOCK_INDEX, _("_Channels"), "<alt>C", NULL, G_CALLBACK (on_network_channels_activate) },
 	{ "NetworkUsers", NULL, _("_Users"), "<alt>U", NULL, G_CALLBACK (on_network_users_activate) },
 
 	/* Discussion menu */
 	{ "DiscussionSave", GTK_STOCK_SAVE, _("_Save Transcript"), "<control>S", NULL, G_CALLBACK (on_discussion_save_activate) },
 	{ "DiscussionLeave", GTK_STOCK_QUIT, _("_Leave"), "", NULL, G_CALLBACK (on_discussion_leave_activate) },
-	{ "DiscussionClose", GTK_STOCK_CLOSE, _("Cl_ose"), "<shift><control>W", NULL, G_CALLBACK (on_discussion_close_activate) },
+	{ "DiscussionClose", GTK_STOCK_CLOSE, _("Cl_ose"), "<control>W", NULL, G_CALLBACK (on_discussion_close_activate) },
 	{ "DiscussionFind", GTK_STOCK_FIND, _("_Find"), "<control>F", NULL, G_CALLBACK (on_discussion_find_activate) },
 	{ "DiscussionFindNext", NULL, _("Find Ne_xt"), "<control>G", NULL, G_CALLBACK (on_discussion_find_next_activate) },
 	{ "DiscussionClearWindow", GTK_STOCK_CLEAR, _("_Clear Window"), "<control>L", NULL, G_CALLBACK (on_discussion_clear_window_activate) },
@@ -160,13 +151,11 @@ static GtkActionEntry action_entries [] = {
 	{ "GoNextDiscussion", GTK_STOCK_GO_FORWARD, _("_Next Discussion"), "<alt>Down", NULL, G_CALLBACK (on_go_next_discussion_activate) },
 
 	/* Help menu */
-#if (GTK_CHECK_VERSION(2,5,0))
 	{ "HelpAbout", GTK_STOCK_ABOUT, _("_About"), "", NULL, G_CALLBACK (on_help_about_activate) },
-#else
-	{ "HelpAbout", NULL, _("_About"), "", NULL, G_CALLBACK (on_help_about_activate) },
-#endif
 };
 
+static GCompletion *command_completion;
+
 /*
  * FIXME: for the love of god make this more efficient
  */
@@ -276,6 +265,7 @@ setup_menu_item (GConfClient *client, GtkActionEntry *entry)
 	if (e == NULL) {
 		/* no GConf data, so let's use the hardcoded values */
 		gtk_action_group_add_actions (gui.action_group, entry, 1, NULL);
+		g_free (key_string);
 		return;
 	}
 
@@ -290,6 +280,7 @@ setup_menu_item (GConfClient *client, GtkActionEntry *entry)
 	if (entry->callback != NULL)
 		g_signal_connect (action, "activate", G_CALLBACK (entry->callback), gui.xml);
 
+	g_object_unref (action);
 	g_free (key_string);
 	gconf_entry_free (e);
 }
@@ -314,8 +305,7 @@ close_find_button (GtkWidget *button, gpointer data)
 	GtkWidget *widget;
 	widget = glade_xml_get_widget (gui.xml, "find hbox");
 	gtk_widget_hide (widget);
-	widget = glade_xml_get_widget (gui.xml, "text entry");
-	gtk_widget_grab_focus (widget);
+	gtk_widget_grab_focus (gui.text_entry);
 }
 
 static gboolean
@@ -340,11 +330,45 @@ clear_find (GtkWidget *entry, gpointer data)
 	last_search_position = NULL;
 }
 
+#ifdef HAVE_LIBSEXY
+static gboolean
+spell_check (SexySpellEntry *entry, gchar *text, gpointer data)
+{
+	GtkTreeModel *store = GTK_TREE_MODEL (userlist_get_store (u, gui.current_session));
+	GtkTreeIter iter;
+
+	if (gtk_tree_model_get_iter_first (store, &iter) == FALSE)
+		return TRUE;
+	do {
+		gchar *nick;
+		gboolean match = FALSE;
+
+		gtk_tree_model_get (store, &iter, 1, &nick, -1);
+		if (strncmp (text, nick, strlen (nick)) == 0)
+			match = TRUE;
+
+		g_free (nick);
+		if (match)
+			return FALSE;
+	} while (gtk_tree_model_iter_next (store, &iter));
+
+	return TRUE;
+}
+
+static void
+url_activated (GtkWidget *url_label, const char *url, gpointer data)
+{
+	fe_open_url (url);
+}
+#endif
+
 void
 initialize_main_window ()
 {
-	GtkWidget *entry, *topicbox, *close, *menu_vbox, *widget;
+	GtkWidget *entrybox, *topicbox, *close, *menu_vbox, *widget;
 	GError *error = NULL;
+	GList *tmp = NULL;
+	int i;
 
 	gui.main_window = GNOME_APP (glade_xml_get_widget (gui.xml, "xchat-gnome"));
 	g_signal_connect (G_OBJECT (gui.main_window), "delete-event",
@@ -379,20 +403,35 @@ initialize_main_window ()
 	/* hook up accelerators */
 	gtk_window_add_accel_group (GTK_WINDOW (gui.main_window), gtk_ui_manager_get_accel_group (gui.manager));
 
-	entry = glade_xml_get_widget (gui.xml, "text entry");
-	g_signal_connect (G_OBJECT (entry), "activate", G_CALLBACK (on_text_entry_activate), NULL);
-	g_signal_connect_after (G_OBJECT (entry), "key_press_event", G_CALLBACK (on_text_entry_key), NULL);
-	g_signal_connect (G_OBJECT (entry), "populate-popup", G_CALLBACK (entry_context), NULL);
+	entrybox = glade_xml_get_widget (gui.xml, "entry hbox");
+#ifdef HAVE_LIBSEXY
+	gui.text_entry = sexy_spell_entry_new ();
+	g_signal_connect_after (G_OBJECT (gui.text_entry), "word-check", G_CALLBACK (spell_check), NULL);
+#else
+	gui.text_entry = gtk_entry_new ();
+#endif
+	gtk_box_pack_start (GTK_BOX (entrybox), gui.text_entry, TRUE, TRUE, 0);
+	gtk_widget_show (gui.text_entry);
+	g_signal_connect (G_OBJECT (gui.text_entry), "activate", G_CALLBACK (on_text_entry_activate), NULL);
+	g_signal_connect_after (G_OBJECT (gui.text_entry), "key_press_event", G_CALLBACK (on_text_entry_key), NULL);
+	g_signal_connect (G_OBJECT (gui.text_entry), "populate-popup", G_CALLBACK (entry_context), NULL);
 
 	close = glade_xml_get_widget (gui.xml, "close discussion");
 	g_signal_connect (G_OBJECT (close), "clicked", G_CALLBACK (on_discussion_close_activate), NULL);
 	topicbox = glade_xml_get_widget (gui.xml, "topic hbox");
-	gui.topic_label = GTK_LABEL (gtk_label_new(""));
-	gtk_widget_show (GTK_WIDGET (gui.topic_label));
-	gtk_box_pack_start (GTK_BOX (topicbox), GTK_WIDGET (gui.topic_label), TRUE, TRUE, 0);
-	gtk_box_reorder_child (GTK_BOX (topicbox), GTK_WIDGET (gui.topic_label), 1);
-	gtk_label_set_selectable (gui.topic_label, TRUE);
-#if (GTK_CHECK_VERSION(2,5,0))
+
+#ifdef HAVE_LIBSEXY
+	gui.topic_label = sexy_url_label_new ();
+	g_signal_connect (gui.topic_label, "url_activated", G_CALLBACK (url_activated), NULL);
+#else
+	gui.topic_label = gtk_label_new("");
+#endif
+	gtk_widget_show (gui.topic_label);
+
+	gtk_box_pack_start (GTK_BOX (topicbox), gui.topic_label, TRUE, TRUE, 0);
+	gtk_box_reorder_child (GTK_BOX (topicbox), gui.topic_label, 1);
+	gtk_label_set_selectable (GTK_LABEL (gui.topic_label), TRUE);
+
 	gui.topic_expander = GTK_EXPANDER (gtk_expander_new (NULL));
 	gtk_widget_show (GTK_WIDGET (gui.topic_expander));
 	gtk_box_pack_start (GTK_BOX (topicbox), GTK_WIDGET (gui.topic_expander), FALSE, TRUE, 0);
@@ -400,10 +439,7 @@ initialize_main_window ()
 	gtk_expander_set_expanded (GTK_EXPANDER (gui.topic_expander), FALSE);
 	gtk_expander_set_use_markup (gui.topic_expander, TRUE);
 	g_signal_connect (G_OBJECT (gui.topic_expander), "activate", G_CALLBACK (on_expand_topic), NULL);
-	gtk_label_set_ellipsize (gui.topic_label, PANGO_ELLIPSIZE_END);
-#else
-	gtk_label_set_line_wrap (gui.topic_label, TRUE);
-#endif
+	gtk_label_set_ellipsize (GTK_LABEL (gui.topic_label), PANGO_ELLIPSIZE_END);
 
 	/* Hook up accelerators for pgup/pgdn */
 	{
@@ -462,25 +498,36 @@ initialize_main_window ()
 		g_closure_unref (closure);
 
 		/* alt-+ */
-		closure = g_cclosure_new (G_CALLBACK (on_discussion_plus_activate), NULL, NULL);
+		closure = g_cclosure_new (G_CALLBACK (on_go_next_discussion_activate), NULL, NULL);
 		gtk_accel_group_connect (discussion_accel, gdk_keyval_from_name ("equal"), GDK_MOD1_MASK, GTK_ACCEL_VISIBLE, closure);
 		g_closure_unref (closure);
 
 		/* alt-- */
-		closure = g_cclosure_new (G_CALLBACK (on_discussion_minus_activate), NULL, NULL);
+		closure = g_cclosure_new (G_CALLBACK (on_go_previous_discussion_activate), NULL, NULL);
 		gtk_accel_group_connect (discussion_accel, gdk_keyval_from_name ("minus"), GDK_MOD1_MASK, GTK_ACCEL_VISIBLE, closure);
 		g_closure_unref (closure);
 
+		/* We've had a couple of calls to hook up Ctrl-Alt-PgUp and
+		 * Ctrl-Alt-PgDown to discussion navigation. As far as I can
+		 * tell this is not HIG compliant, but for the time being we'll
+		 * put it in. It's easy enough to delete it later.
+		 */
+		closure = g_cclosure_new (G_CALLBACK (on_go_next_discussion_activate), NULL, NULL);
+		gtk_accel_group_connect (discussion_accel, GDK_Page_Down,
+				GDK_MOD1_MASK | GDK_CONTROL_MASK , GTK_ACCEL_VISIBLE, closure);
+
+		g_closure_unref (closure);
+
+		closure = g_cclosure_new (G_CALLBACK (on_go_previous_discussion_activate), NULL, NULL);
+		gtk_accel_group_connect (discussion_accel, GDK_Page_Up,
+				GDK_MOD1_MASK | GDK_CONTROL_MASK, GTK_ACCEL_VISIBLE, closure);
+
+		g_closure_unref (closure);
+
 		/* Add the accelgroup to the main window. */
 		gtk_window_add_accel_group (GTK_WINDOW (gui.main_window), discussion_accel);
 	}
 
-#ifdef HAVE_GTKSPELL
-#if 0
-	gtkspell_new_attach(GTK_TEXT_VIEW(entry), NULL, NULL);
-#endif
-#endif
-
 	/* setup find stuff */
 	widget = glade_xml_get_widget (gui.xml, "find entry");
 	g_signal_connect (G_OBJECT (widget), "activate", G_CALLBACK (find_next), NULL);
@@ -488,6 +535,22 @@ initialize_main_window ()
 	g_signal_connect (G_OBJECT (widget), "changed", G_CALLBACK (clear_find), NULL);
 	widget = glade_xml_get_widget (gui.xml, "find close button");
 	g_signal_connect (G_OBJECT (widget), "clicked", G_CALLBACK (close_find_button), NULL);
+
+	/* connect nickname button */
+	widget = glade_xml_get_widget (gui.xml, "nickname");
+	gtk_button_set_use_underline (GTK_BUTTON (widget), FALSE);
+	g_signal_connect (G_OBJECT (widget), "clicked", G_CALLBACK (on_nickname_clicked), NULL);
+
+	/* initialize command completion */
+	/* FIXME - would be nice to find somewhere else for this */
+	command_completion = g_completion_new (NULL);
+	g_completion_add_items (command_completion, command_list);
+	for (i = 0; xc_cmds[i].name != NULL; i++) {
+		tmp = g_list_prepend (tmp, xc_cmds[i].name);
+	}
+	g_list_reverse (tmp);
+	g_completion_add_items (command_completion, tmp);
+	g_list_free (tmp);
 }
 
 void
@@ -522,12 +585,14 @@ run_main_window ()
 	gtk_widget_show (GTK_WIDGET (gui.main_window));
 
 	/* Temporarily disable menu items */
-	widget = gtk_ui_manager_get_widget (gui.manager, "/ui/menubar/InsertMenu/InsertColorCodeItem");
-	gtk_widget_set_sensitive (widget, FALSE);
 	widget = gtk_ui_manager_get_widget (gui.manager, "/ui/menubar/NetworkMenu/NetworkInformationItem");
 	gtk_widget_set_sensitive (widget, FALSE);
+	widget = gtk_ui_manager_get_widget (gui.manager, "/ui/menubar/NetworkMenu/NetworkUsersItem");
+	gtk_widget_set_sensitive (widget, FALSE);
 	widget = gtk_ui_manager_get_widget (gui.manager, "/ui/menubar/DiscussionMenu/DiscussionSaveItem");
 	gtk_widget_set_sensitive (widget, FALSE);
+	widget = gtk_ui_manager_get_widget (gui.manager, "/ui/menubar/DiscussionMenu/DiscussionBansItem");
+	gtk_widget_set_sensitive (widget, FALSE);
 }
 
 void
@@ -561,6 +626,7 @@ static void
 on_main_window_close (GtkWidget *widget, GdkEvent *event, gpointer data)
 {
 	gui.quit = TRUE;
+	gtk_widget_hide (GTK_WIDGET (gui.dcc));
 	xchat_exit ();
 }
 
@@ -568,6 +634,7 @@ static void
 on_irc_quit_activate (GtkAction *action, gpointer data)
 {
 	gtk_widget_hide (GTK_WIDGET (gui.main_window));
+	gtk_widget_hide (GTK_WIDGET (gui.dcc));
 	gui.quit = TRUE;
 	xchat_exit ();
 }
@@ -575,29 +642,25 @@ on_irc_quit_activate (GtkAction *action, gpointer data)
 static void
 on_edit_cut_activate (GtkAction *action, gpointer data)
 {
-	GtkWidget *text_entry = glade_xml_get_widget(gui.xml, "text entry");
-	gtk_editable_cut_clipboard(GTK_EDITABLE(text_entry));
+	gtk_editable_cut_clipboard(GTK_EDITABLE(gui.text_entry));
 }
 
 static void
 on_edit_copy_activate (GtkAction *action, gpointer data)
 {
-	GtkWidget *text_entry = glade_xml_get_widget(gui.xml, "text entry");
-	gtk_editable_copy_clipboard(GTK_EDITABLE(text_entry));
+	gtk_editable_copy_clipboard(GTK_EDITABLE(gui.text_entry));
 }
 
 static void
 on_edit_paste_activate (GtkAction *action, gpointer data)
 {
-	GtkWidget *text_entry = glade_xml_get_widget(gui.xml, "text entry");
-	gtk_editable_paste_clipboard(GTK_EDITABLE(text_entry));
+	gtk_editable_paste_clipboard(GTK_EDITABLE(gui.text_entry));
 }
 
 static void
 on_edit_clear_activate (GtkAction *action, gpointer data)
 {
-	GtkWidget *text_entry = glade_xml_get_widget(gui.xml, "text entry");
-	gtk_editable_delete_selection(GTK_EDITABLE(text_entry));
+	gtk_editable_delete_selection(GTK_EDITABLE(gui.text_entry));
 }
 
 static void
@@ -610,12 +673,6 @@ on_edit_preferences_activate (GtkAction *action, gpointer data)
 	gtk_widget_destroy (GTK_WIDGET (dialog));
 }
 
-static void
-on_insert_color_code_activate (GtkAction *action, gpointer data)
-{
-	/* FIXME: implement */
-}
-
 static void
 on_network_information_activate (GtkAction *action, gpointer data)
 {
@@ -632,15 +689,19 @@ static void
 on_network_disconnect_activate (GtkAction *action, gpointer data)
 {
 	session *s = gui.current_session;
-	s->server->disconnect (s, TRUE, -1);
+	if (s)
+		s->server->disconnect (s, TRUE, -1);
 }
 
 static void
 on_network_close_activate (GtkAction *actoin, gpointer data)
 {
+	GtkTreeIter parent, *iter;
 	session *sess = gui.current_session;
-	GtkTreeIter parent,
-				*iter =	navigation_model_get_unsorted_iter (gui.tree_model, sess);
+	if (sess == NULL)
+		return;
+
+	iter = navigation_model_get_unsorted_iter (gui.tree_model, sess);
 
 	sess->server->disconnect (sess, TRUE, -1);
 
@@ -656,6 +717,7 @@ on_network_close_activate (GtkAction *actoin, gpointer data)
 static void
 on_irc_downloads_activate (GtkAction *action, gpointer data)
 {
+	gtk_window_present (GTK_WINDOW (gui.dcc));
 }
 
 static void
@@ -698,6 +760,8 @@ static void
 on_discussion_close_activate (GtkAction *action, gpointer data)
 {
 	session *s = gui.current_session;
+	if (s == NULL)
+		return;
 /*	navigation_tree_select_next_channel (gui.server_tree, TRUE);*/
 	if (s->type == SESS_CHANNEL) {
 		gchar *text;
@@ -711,7 +775,7 @@ on_discussion_close_activate (GtkAction *action, gpointer data)
 		g_object_unref (client);
 		g_free (text);
 	}
-	navigation_tree_remove_channel (gui.server_tree, s);
+	fe_close_window (s);
 	text_gui_remove_text_buffer (s);
 }
 
@@ -760,13 +824,13 @@ on_go_next_network_activate(GtkAction *action, gpointer data)
 static void
 on_go_previous_discussion_activate (GtkAction *action, gpointer data)
 {
-	navigation_tree_select_prev_channel (gui.server_tree, TRUE);
+	navigation_tree_select_prev_channel (gui.server_tree);
 }
 
 static void
 on_go_next_discussion_activate (GtkAction *action, gpointer data)
 {
-	navigation_tree_select_next_channel (gui.server_tree, TRUE);
+	navigation_tree_select_next_channel (gui.server_tree);
 }
 
 static void
@@ -775,18 +839,6 @@ on_discussion_jump_activate (GtkAccelGroup *accelgroup, GObject *arg1, guint arg
 	navigation_tree_select_nth_channel (gui.server_tree, GPOINTER_TO_INT (data));
 }
 
-static void
-on_discussion_plus_activate(GtkAccelGroup *accelgroup, GObject *arg1, guint arg2, GdkModifierType arg3, gpointer data)
-{
-	navigation_tree_select_next_channel (gui.server_tree, FALSE);
-}
-
-static void
-on_discussion_minus_activate (GtkAccelGroup *accelgroup, GObject *arg1, guint arg2, GdkModifierType arg3, gpointer data)
-{
-	navigation_tree_select_prev_channel (gui.server_tree, FALSE);
-}
-
 static void
 on_pgup (GtkAccelGroup *accelgroup, GObject *arg1, guint arg2, GdkModifierType arg3, gpointer data)
 {
@@ -834,10 +886,55 @@ on_text_entry_activate (GtkWidget *widget, gpointer data)
 {
 	char *entry_text = g_strdup (gtk_entry_get_text (GTK_ENTRY (widget)));
 	gtk_entry_set_text (GTK_ENTRY (widget), "");
-	handle_multiline (gui.current_session, (char *) entry_text, TRUE, FALSE);
+	if (gui.current_session != NULL)
+		handle_multiline (gui.current_session, (char *) entry_text, TRUE, FALSE);
 	g_free (entry_text);
 }
 
+static void
+nickname_get_str_response (GtkDialog *dialog, gint arg1, gpointer entry)
+{
+	gchar *text, *buf;
+
+	text = (gchar *) gtk_entry_get_text (GTK_ENTRY (entry));
+
+	if (arg1 == GTK_RESPONSE_ACCEPT || arg1 == GTK_RESPONSE_OK) {
+		buf = g_strdup_printf ("nick %s", text);
+		handle_command (current_sess, buf, FALSE);
+		g_free (buf);
+	}
+	gtk_widget_hide_all (GTK_WIDGET (dialog));
+}
+
+static void
+nickname_str_enter (GtkWidget *entry, GtkWidget *dialog)
+{
+	gtk_dialog_response (GTK_DIALOG (dialog), GTK_RESPONSE_ACCEPT);
+}
+
+static void
+on_nickname_clicked (GtkButton *widget, gpointer user_data)
+{
+	GtkWidget *dialog;
+	GtkWidget *entry;
+
+	if (!current_sess)
+		return;
+
+	dialog = glade_xml_get_widget (gui.xml, "nickname dialog");
+	entry = glade_xml_get_widget (gui.xml, "nickname dialog entry");
+
+	gtk_entry_set_text (GTK_ENTRY (entry), current_sess->server->nick);
+	g_signal_connect (G_OBJECT (entry), "activate",
+						 	G_CALLBACK (nickname_str_enter),
+						 	dialog);
+	g_signal_connect (G_OBJECT (dialog), "response",
+						   G_CALLBACK (nickname_get_str_response),
+						   entry);
+
+	gtk_widget_show_all (dialog);
+}
+
 static void
 history_key_down (GtkEntry *entry)
 {
@@ -860,6 +957,74 @@ history_key_up (GtkEntry *entry)
 	}
 }
 
+static gboolean
+tab_complete_command (GtkEntry *entry)
+{
+	int cursor, length, pos;
+	char *prefix, *new_prefix, *text, *npt, *printtext;
+	GList *options, *list;
+	session_gui *tgui;
+
+	cursor = gtk_editable_get_position (GTK_EDITABLE (entry));
+	prefix = g_new0 (char, cursor);
+	text = gtk_entry_get_text (entry);
+	strncpy (prefix, &text[1], cursor - 1);
+	length = strlen (text);
+
+	options = g_completion_complete (command_completion, prefix, &new_prefix);
+
+	if (g_list_length (options) == 0) {
+		/* no matches */
+		g_free (prefix);
+		return TRUE;
+	}
+
+	if (g_list_length (options) == 1) {
+		/* one match */
+
+		if (length - cursor == 0) {
+			/* at the end of the entry, just insert */
+
+			npt = g_strdup_printf ("/%s ", (char *) options->data);
+			pos = strlen (npt);
+		} else {
+			npt = g_strdup_printf ("/%s %s", (char *) options->data, &text[cursor]);
+			pos = strlen ((char *) options->data) + 2;
+		}
+		gtk_entry_set_text (entry, npt);
+		gtk_editable_set_position (GTK_EDITABLE (entry), pos);
+		g_free (npt);
+		g_free (prefix);
+		return TRUE;
+	} else {
+		/* more than one match - print a list of options
+		 * to the window and update the prefix
+		 */
+		list = options;
+		printtext = g_strdup ((char *) list->data);
+		for (list = g_list_next (list); list; list = g_list_next (list)) {
+			npt = g_strdup_printf ("%s %s", printtext, (char *) list->data);
+			g_free (printtext);
+			printtext = npt;
+		}
+		tgui = (session_gui *) gui.current_session->gui;
+		text_gui_print (tgui->buffer, (guchar *) printtext, TRUE);
+		g_free (printtext);
+		if (strcasecmp (prefix, new_prefix) != 0) {
+			/* insert the new prefix into the entry */
+			npt = g_strdup_printf ("/%s%s", new_prefix, &text[cursor]);
+			gtk_entry_set_text (entry, npt);
+			g_free (npt);
+			gtk_editable_set_position (GTK_EDITABLE (entry), strlen (new_prefix));
+		}
+		g_free (prefix);
+		g_free (npt);
+		return TRUE;
+	}
+
+	return TRUE;
+}
+
 static gboolean
 tab_complete_nickname (GtkEntry *entry, int start)
 {
@@ -929,7 +1094,7 @@ tab_complete_nickname (GtkEntry *entry, int start)
 		 */
 		list = options;
 		printtext = g_strdup ((char *) list->data);
-		for (list = list->next; list; list = list->next) {
+		for (list = g_list_next (list); list; list = g_list_next (list)) {
 			npt = g_strdup_printf ("%s %s", printtext, (char *) list->data);
 			g_free (printtext);
 			printtext = npt;
@@ -943,12 +1108,13 @@ tab_complete_nickname (GtkEntry *entry, int start)
 			npt = g_strdup_printf ("%s%s%s", text, new_prefix, &text[cursor]);
 			gtk_entry_set_text (entry, npt);
 			g_free (npt);
-			gtk_editable_set_position (GTK_EDITABLE(entry), start + strlen(new_prefix));
+			gtk_editable_set_position (GTK_EDITABLE (entry), start + strlen (new_prefix));
 		}
-		g_free(text);
+		g_free (text);
 		g_free (prefix);
 		return TRUE;
 	}
+	return TRUE;
 }
 
 static gboolean
@@ -981,7 +1147,7 @@ tab_complete (GtkEntry *entry)
 
 		/* check if we can match a command */
 		if (start == 0 && text[0] == '/') {
-		    /* TODO: Something? */
+			return tab_complete_command (entry);
 		}
 
 		/* check if we can match a nickname */
@@ -1042,11 +1208,13 @@ on_discussion_topic_change_activate (GtkButton *widget, gpointer data)
 	GladeXML *xml = NULL;
 	GtkWidget *dialog;
 	GtkWidget *entry;
-	GtkWidget *text_entry;
 	gint response;
 	GtkTextBuffer *buffer;
 	gchar *title;
 
+	if (gui.current_session == NULL)
+		return;
+
 	if (g_file_test ("topic-change.glade", G_FILE_TEST_EXISTS))
 		xml = glade_xml_new ("topic-change.glade", NULL, NULL);
 	if (!xml)
@@ -1087,8 +1255,7 @@ on_discussion_topic_change_activate (GtkButton *widget, gpointer data)
 	g_object_unref (xml);
 
 	/* send focus back to the text entry */
-	text_entry = glade_xml_get_widget (gui.xml, "text entry");
-	gtk_widget_grab_focus (text_entry);
+	gtk_widget_grab_focus (gui.text_entry);
 }
 
 void
@@ -1131,11 +1298,10 @@ static void
 color_code_activate (GtkMenuItem *item, gpointer data)
 {
 	int color = (int) data;
-	GtkWidget *entry = glade_xml_get_widget (gui.xml, "text entry");
 	char *code = g_strdup_printf ("%%C%d", color);
-	int position = gtk_editable_get_position (GTK_EDITABLE (entry));
-	gtk_editable_insert_text (GTK_EDITABLE (entry), code, strlen (code), &position);
-	gtk_editable_set_position (GTK_EDITABLE (entry), position + strlen (code));
+	int position = gtk_editable_get_position (GTK_EDITABLE (gui.text_entry));
+	gtk_editable_insert_text (GTK_EDITABLE (gui.text_entry), code, strlen (code), &position);
+	gtk_editable_set_position (GTK_EDITABLE (gui.text_entry), position + strlen (code));
 	g_free (code);
 }
 
@@ -1220,20 +1386,14 @@ entry_context (GtkEntry *entry, GtkMenu *menu, gpointer user_data)
 	gtk_widget_show_all (submenu);
 }
 
-#if (GTK_CHECK_VERSION(2,5,0))
 static void
 on_expand_topic (GtkExpander *expander, gpointer data)
 {
 	if (gtk_expander_get_expanded (gui.topic_expander)) {
-		gtk_label_set_ellipsize(gui.topic_label, PANGO_ELLIPSIZE_END);
-		gtk_label_set_line_wrap (gui.topic_label, FALSE);
+		gtk_label_set_ellipsize(GTK_LABEL (gui.topic_label), PANGO_ELLIPSIZE_END);
+		gtk_label_set_line_wrap (GTK_LABEL (gui.topic_label), FALSE);
 	} else {
-		gtk_label_set_ellipsize(gui.topic_label, PANGO_ELLIPSIZE_NONE);
-		gtk_label_set_line_wrap (gui.topic_label, TRUE);
+		gtk_label_set_ellipsize(GTK_LABEL (gui.topic_label), PANGO_ELLIPSIZE_NONE);
+		gtk_label_set_line_wrap (GTK_LABEL (gui.topic_label), TRUE);
 	}
 }
-#endif
-
-/*** The End ***/
-/* vim: ts=3:sw=3
- */
